import datetime as dt
from typing import ClassVar, List, Optional, Union, cast

from clabe import Clabe
from cuenca_validations.types import (
    PaymentCardNumber,
    StrictPositiveInt,
    TransferNetwork,
    TransferQuery,
)
from cuenca_validations.typing import DictStrAny
from pydantic import BaseModel, StrictStr
from pydantic.dataclasses import dataclass
from requests import HTTPError

from ..exc import CuencaException
from .accounts import Account
from .base import Creatable, Transaction
from .resources import retrieve_uri


class TransferRequest(BaseModel):
    recipient_name: StrictStr
    account_number: Union[Clabe, PaymentCardNumber]
    amount: StrictPositiveInt  # in centavos
    descriptor: StrictStr  # how it'll appear for the recipient
    idempotency_key: str  # must be unique for each transfer


@dataclass
class Transfer(Transaction, Creatable):
    _resource: ClassVar = 'transfers'
    _query_params: ClassVar = TransferQuery

    updated_at: dt.datetime
    recipient_name: str
    account_number: str
    idempotency_key: str
    network: TransferNetwork
    tracking_key: Optional[str]  # clave rastreo if network is SPEI
    destination_uri: Optional[str]  # defined after confirmation of receipt

    @property  # type: ignore
    def destination(self) -> Optional[Account]:
        if self.destination_uri is None:
            acct = None
        else:
            acct = cast(Account, retrieve_uri(self.destination_uri))
        return acct

    @classmethod
    def create(
        cls,
        account_number: str,
        amount: int,
        descriptor: str,
        recipient_name: str,
        idempotency_key: Optional[str] = None,
    ) -> 'Transfer':
        """
        :param account_number: CLABE
        :param amount: needs to be in centavos (not pesos)
        :param descriptor: how it'll appear for the recipient
        :param recipient_name: name of recipient
        :param idempotency_key: must be unique for each transfer to avoid
            duplicates
        :return: Transfer object

        The recommended idempotency_key scheme:
        1. create a transfer entry in your own database with the status
            created
        2. call this method with the unique id from your database as the
            idempotency_key
        3. update your database with the status created or submitted after
            receiving a response from this method
        """
        if not idempotency_key:
            idempotency_key = cls._gen_idempotency_key(account_number, amount)
        req = TransferRequest(
            account_number=account_number,
            amount=amount,
            descriptor=descriptor,
            recipient_name=recipient_name,
            idempotency_key=idempotency_key,
        )
        return cast('Transfer', cls._create(**req.dict()))

    @classmethod
    def create_many(cls, requests: List[TransferRequest]) -> DictStrAny:
        transfers: DictStrAny = dict(submitted=[], errors=[])
        for req in requests:
            try:
                transfer = cls._create(**req.dict())
            except (CuencaException, HTTPError) as e:
                transfers['errors'].append(dict(request=req, error=e))
            else:
                transfers['submitted'].append(cast('Transfer', transfer))
        return transfers

    @staticmethod
    def _gen_idempotency_key(account_number: str, amount: int) -> str:
        """
        We *strongly* recommend using your own internal database id as the
        idempotency_key, but this provides some level of protection against
        submitting duplicate transfers
        """
        return f'{dt.datetime.utcnow().date()}:{account_number}:{amount}'



from typing import ClassVar, cast

from cuenca_validations.types import EntryType, RelatedTransaction
from pydantic.dataclasses import dataclass

from cuenca import resources

from .base import Queryable, Retrievable
from .resources import retrieve_uri


@dataclass
class BalanceEntry(Retrievable, Queryable):
    _resource: ClassVar = 'balance_entries'

    amount: int  # negative in the case of a debit
    descriptor: str
    name: str
    rolling_balance: int
    type: EntryType
    related_transaction_uri: RelatedTransaction

    @property  # type: ignore
    def related_transaction(self):
        rt = self.related_transaction_uri
        resource = getattr(resources, rt.get_model())
        return cast(resource, retrieve_uri(rt)) if resource else None



from typing import ClassVar, Optional, cast

from cuenca_validations.types import CardStatus, CardType
from cuenca_validations.types.queries import CardQuery
from cuenca_validations.types.requests import CardRequest, CardUpdateRequest
from pydantic.dataclasses import dataclass

from cuenca.resources.base import Creatable, Queryable, Retrievable, Updateable

from ..http import session


@dataclass
class Card(Retrievable, Queryable, Creatable, Updateable):
    _resource: ClassVar = 'cards'
    _query_params: ClassVar = CardQuery

    user_id: str
    ledger_account_id: str
    number: str
    exp_month: int
    exp_year: int
    cvv2: str
    type: CardType
    status: CardStatus

    @classmethod
    def create(cls, ledger_account_id: str, user_id: str) -> 'Card':
        """
        Assigns user_id and ledger_account_id to a existing card

        :param ledger_account_id: associated ledger account id
        :param user_id: associated user id
        :return: New assigned card
        """
        req = CardRequest(ledger_account_id=ledger_account_id, user_id=user_id)
        return cast('Card', cls._create(**req.dict()))

    @classmethod
    def update(
        cls,
        card_id: str,
        user_id: Optional[str] = None,
        ledger_account_id: Optional[str] = None,
        status: Optional[CardStatus] = None,
    ):
        """
        Updates card properties that are not sensitive or fixed data. It allows
        reconfigure properties like status, and manufacturer.

        :param card_id: existing card_id
        :param user_id: owner user id
        :param ledger_account_id: owner ledger account
        :param status:
        :return: Updated card object
        """
        req = CardUpdateRequest(
            user_id=user_id, ledger_account_id=ledger_account_id, status=status
        )
        resp = cls._update(card_id, **req.dict(exclude_none=True))
        return cast('Card', resp)

    @classmethod
    def deactivate(cls, card_id: str) -> 'Card':
        """
        Deactivates a card
        """
        url = f'{cls._resource}/{card_id}'
        resp = session.delete(url)
        return cast('Card', cls._from_dict(resp))



import pytest

from cuenca import BalanceEntry
from cuenca.exc import CuencaResponseException


@pytest.mark.vcr
def test_balance_entry_retrieve():
    id_entry = 'TV01'
    balance_entry: BalanceEntry = BalanceEntry.retrieve(id_entry)
    assert balance_entry.id == id_entry
    txn = balance_entry.related_transaction
    assert (
        balance_entry.related_transaction_uri == f'/{txn._resource}/{txn.id}'
    )


@pytest.mark.vcr
def test_balance_entry_related_transaction_not_exist():
    id_entry = 'TV02'
    balance_entry: BalanceEntry = BalanceEntry.retrieve(id_entry)
    assert balance_entry.id == id_entry
    with pytest.raises(CuencaResponseException):
        balance_entry.related_transaction
